from pathlib import Path

import pytest

from zhinst.toolkit.driver.modules.device_settings_module import DeviceSettingsModule


@pytest.fixture
def device_settings_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_device_settings_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.deviceSettings.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    return DeviceSettingsModule(mock_connection.return_value.deviceSettings(), session)


def test_repr(device_settings_module):
    assert "MagicMock(DataServerSession(localhost:8004))" in repr(
        device_settings_module,
    )


def test_load_from_file(device_settings_module):
    device_settings_module.raw_module.getInt.return_value = 1
    device_settings_module.raw_module.get.return_value = {
        "/finished": {"timestamp": [0], "value": [1]},
    }

    device_settings_module.load_from_file(Path("/path/to/file.json"), "dev1234")
    device_settings_module.raw_module.set.assert_any_call("/device", "dev1234")
    device_settings_module.raw_module.set.assert_any_call("/filename", "file")
    device_settings_module.raw_module.set.assert_any_call("/path", "/path/to")
    device_settings_module.raw_module.set.assert_any_call("/command", "load")
    device_settings_module.raw_module.execute.assert_called_once()


def test_load_from_file_timeout(device_settings_module):
    device_settings_module.raw_module.getInt.return_value = 0
    device_settings_module.raw_module.get.return_value = {
        "/finished": {"timestamp": [0], "value": [0]},
    }

    with pytest.raises(TimeoutError):
        device_settings_module.load_from_file(
            Path("/path/to/file.json"),
            "dev1234",
            timeout=0.01,
        )


def test_save_to_file(device_settings_module):
    device_settings_module.raw_module.getInt.return_value = 1
    device_settings_module.raw_module.get.return_value = {
        "/finished": {"timestamp": [0], "value": [1]},
    }

    device_settings_module.save_to_file(Path("/path/to/file.json"), "dev1234")
    device_settings_module.raw_module.set.assert_any_call("/device", "dev1234")
    device_settings_module.raw_module.set.assert_any_call("/filename", "file")
    device_settings_module.raw_module.set.assert_any_call("/path", "/path/to")
    device_settings_module.raw_module.set.assert_any_call("/command", "save")
    device_settings_module.raw_module.execute.assert_called_once()


def test_save_to_file_timeout(device_settings_module):
    device_settings_module.raw_module.getInt.return_value = 0
    device_settings_module.raw_module.get.return_value = {
        "/finished": {"timestamp": [0], "value": [0]},
    }

    with pytest.raises(TimeoutError):
        device_settings_module.save_to_file(
            Path("/path/to/file.json"),
            "dev1234",
            timeout=0.01,
        )


def test_read(device_settings_module):
    device_settings_module.raw_module.read.return_value = {"/device": "test"}
    result = device_settings_module.read()
    assert device_settings_module.device in result
    assert result[device_settings_module.device] == "test"



# Copyright (C) 2020 Zurich Instruments
#
# This software may be modified and distributed under the terms
# of the MIT license. See the LICENSE file for details.
"""The Zurich Instruments Toolkit (zhinst-toolkit).

This package is a collection of Python tools for high level device
control. Based on the native interface to Zurich Instruments LabOne,
they offer an easy and user-friendly way to control Zurich Instruments
devices. It  is tailored to control multiple instruments together,
especially for device management and multiple AWG distributed control.
"""
from zhinst.toolkit.command_table import CommandTable
from zhinst.toolkit.driver.modules.pid_advisor_module import PIDMode
from zhinst.toolkit.interface import AveragingMode, SHFQAChannelMode
from zhinst.toolkit.sequence import Sequence
from zhinst.toolkit.session import PollFlags, Session
from zhinst.toolkit.waveform import Waveforms

try:
    from zhinst.toolkit._version import version as __version__
except ModuleNotFoundError:  # pragma: no cover
    pass

__all__ = [
    "AveragingMode",
    "CommandTable",
    "PIDMode",
    "PollFlags",
    "SHFQAChannelMode",
    "Sequence",
    "Session",
    "Waveforms",
    "__version__",
]



import numpy as np
import pytest

from zhinst.toolkit.driver.modules.scope_module import ScopeModule


@pytest.fixture
def scope_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_scope_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.scopeModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    return ScopeModule(mock_connection.return_value.scopeModule(), session)


def test_repr(scope_module):
    assert "MagicMock(DataServerSession(localhost:8004))" in repr(scope_module)


def test_finish(scope_module, mock_connection):
    module_mock = mock_connection.return_value.scopeModule.return_value
    scope_module.finish()
    module_mock.finish.assert_called_with()


def test_progress(scope_module, mock_connection):
    module_mock = mock_connection.return_value.scopeModule.return_value
    module_mock.progress.return_value = 0.5 * np.ones(1)
    assert scope_module.progress() == 0.5


def test_read(scope_module, mock_connection):
    module_mock = mock_connection.return_value.scopeModule.return_value
    module_mock.read.return_value = {"/externalscaling": np.ones(1)}
    result = scope_module.read()
    assert scope_module.externalscaling in result
    assert result[scope_module.externalscaling][0] == 1



"""Zurich Instruments Toolkit (zhinst-toolkit) Command Table Module.

This module provides a `CommandTable` class to create and modify ZI device
command tables.
"""

from __future__ import annotations

import copy
import json
import typing as t

import jsonref
import jsonschema

from zhinst.toolkit.exceptions import ValidationError

# JSON Schema validator for validating the schemes.
JSON_SCHEMA_VALIDATOR = jsonschema.Draft4Validator


def _validate_instance(
    instance: object,
    schema: dict,
    validator=JSON_SCHEMA_VALIDATOR,
) -> None:
    """Validate JSON instance.

    Args:
        instance: Instance to be validated.
        schema: Schema
        validation: Validator

    Raises:
        ValidationError: Validation failed.
    """
    try:
        jsonschema.validate(
            instance=instance,
            schema=schema,
            cls=validator,
        )
    except jsonschema.ValidationError as e:
        raise ValidationError(str(e)) from None


class ParentNode:
    """ParentNode of the command table.

    ParentNode can contain one or multiple arguments and child ParentNodes.
    It offers a dictionary-like object to manipulate command table properties.
    Similar to the device nodes, it supports accessing the properties by attribute.

    Args:
        schema: JSON schema of the node.
        path: Path representation of the node.
        active_validation: Enable active validation.
    """

    def __init__(
        self,
        schema: dict,
        path: tuple[str, ...],
        active_validation: bool = True,
    ):
        self._schema = schema
        self._path = path
        self._childs: dict[t.Union[str, int], t.Any] = {}
        self._active_validation = active_validation

    def __repr__(self) -> str:
        return "/" + "/".join(self._path)

    def _validate_instance(self, instance: object, schema: dict) -> None:
        """Validate JSON instance.

        Args:
            instance: Instance to be validated.
            schema: Schema

        Raises:
            ValidationError: Validation failed.
        """
        if self._active_validation:
            _validate_instance(instance, schema)

    def is_empty(self) -> bool:
        """Check if the Node is empty and has no properties.

        Returns:
            bool: If children exists.
        """
        return not bool(self._childs)


def _change_active_validation(obj: ParentNode, value: bool) -> None:
    """Change object active validation state.

    Args:
        obj: Object
        value: State of validation
    """
    obj._active_validation = value


class ParentEntry(ParentNode):
    """Parent entry of the CommandTable.

    The parent can have both properties and child properties.

    Args:
        schema: JSON schema of the node.
        path: Path representation of the node.
        active_validation: Enable active validation.
    """

    def __init__(
        self,
        schema: dict,
        path: tuple[str, ...],
        active_validation: bool = True,
    ):
        super().__init__(schema, path, active_validation)
        self._attributes = {}
        self._child_props = {}
        self._properties: dict[str, t.Any] = {}
        # ParentEntry does not have json list entries at this moment.
        for name, property_ in schema["properties"].items():
            if "properties" in property_:
                self._child_props[name] = property_
            else:
                self._attributes[name] = property_

    def __contains__(self, k):
        return k in self._child_props or self._attributes.get(k, None) is not None

    def __dir__(self):
        dir_info = set()
        for k in super().__dir__():
            if not k.startswith("_"):
                dir_info.add(k)
        return dir_info.union(set(self._schema["properties"].keys()))

    def __getattr__(self, name: str) -> t.Union[ParentEntry, t.Any]:
        if name.startswith("_"):
            return None
        try:
            return self._childs[name]
        except KeyError:
            if name in self._child_props:
                self._childs[name] = ParentEntry(
                    self._child_props[name],
                    (*self._path, name),
                    self._active_validation,
                )
                return self._childs[name]
            if name in self._attributes:
                return self._properties.get(name, None)
        msg = f"{name}. Available entries: {self._available_attributes()}"
        raise AttributeError(
            msg,
        )

    def __setattr__(self, name: str, value: t.Any):
        if name.startswith("_"):
            super().__setattr__(name, value)
        elif self._attributes and name in self._attributes:
            if value is None:
                self._childs.pop(name, None)
                self._properties.pop(name, None)
            else:
                self._validate_instance(value, self._attributes[name])
                self._childs[name] = value
                self._properties[name] = value
        elif value is None and name in self._childs:
            self._childs.pop(name)
        else:
            msg = f"{name}. Available entries: {self._available_attributes()}"
            raise AttributeError(
                msg,
            )

    def _available_attributes(self) -> list[str]:
        """Available property attributes for the instance."""
        return list(
            self._attributes.keys() if self._attributes else self._child_props.keys(),
        )

    def as_dict(self) -> dict:
        """Return a dictionary presentation of the table node.

        Returns:
            dict: Table node as dictionary.
        """
        result = {}
        for name, child in self._childs.items():
            if isinstance(child, (ParentEntry, HeaderEntry)):
                if not child.is_empty():
                    result[name] = child.as_dict()
            else:
                result[name] = child
        return result

    def info(self, value: t.Optional[str] = None) -> dict:
        """Get info about the property.

        Args:
            value: Info about to specific property. Otherwise
                return info about the whole property.

        Returns:
            Info about the property.
        """
        return self._schema["properties"].get(value, None) if value else self._schema

    def clear(self) -> None:
        """Clear all properties from the object."""
        self._childs = {}
        self._properties = {}


class ListEntry(ParentNode):
    """List entry of a command table.

    Args:
        schema: JSON schema of the node.
        path: Path representation of the node.
        active_validation: Enable active validation.
    """

    def __init__(
        self,
        schema: dict,
        index_schema: dict,
        path: tuple[str, ...],
        active_validation: bool = True,
    ):
        super().__init__(schema, path, active_validation)
        self._min_length = schema["minItems"]
        self._max_length = schema["maxItems"]
        self._index_schema = index_schema

    def __len__(self):
        return len(self._childs)

    def __getitem__(self, number: int) -> ParentEntry:
        try:
            return self._childs[number]
        except KeyError:
            self._validate_instance(number, self._index_schema)
            self._childs[number] = ParentEntry(
                self._schema["items"],
                (*self._path, str(number)),
                self._active_validation,
            )
            return self._childs[number]

    def __delitem__(self, key):
        del self._childs[key]

    @property
    def range(self) -> tuple[int, int]:
        """Get the range for number of minimum and maximum items in the table.

        Returns:
            Range for number of items in the table.
        """
        return (self._min_length, self._max_length)

    def as_list(self) -> list[dict]:
        """Return a list representation of the table.

        Returns:
            List of dictionary representation of entries in the table.
        """
        table = []
        for name, child in self._childs.items():
            if isinstance(child, ParentEntry):
                json_ = child.as_dict()
                if json_:
                    item = {"index": name}
                    item.update(json_)
                    table.append(item)
        return table


class HeaderEntry(ParentEntry):
    """Header entry of a command table.

    Args:
        schema: JSON schema of the node.
        path: Path representation of the node.
        version: JSON schema version
        active_validation: Enable active validation.
    """

    def __init__(
        self,
        schema: dict,
        path: tuple,
        version: t.Optional[str] = None,
        active_validation: bool = True,
    ):
        super().__init__(schema, path, active_validation)
        # L1 22.08 new schema format
        if version:
            self._childs["version"] = version
        else:
            self._childs["version"] = schema["properties"]["version"]["enum"][0]

    @property
    def version(self) -> str:
        """Version of the schema."""
        return self._childs["version"]


def _derefence_json(schema: t.Union[str, dict]) -> t.Any:
    """Dereference JSON schema.

    Args:
        schema: JSON schema as a string or dictionary.

    Returns:
        Dereferenced schema.

    Raises:
        ValueError: Wrong `schema` type.
    """
    if isinstance(schema, str):
        return jsonref.loads(schema, jsonschema=True)
    if isinstance(schema, dict):
        return jsonref.loads(json.dumps(schema), jsonschema=True)
    raise ValueError(schema)


class CommandTable:
    """Representation of a ZI device command table.

    The class provides functionality to create and modify existing command tables.
    The CommandTable can be modified by via ``header`` and ``table`` properties.

    Args:
        json_schema: JSON Schema of the command table.
        active_validation: Active validation of table entries. (default = True)

            Active validation enabled:

            Each time a table entry is accessed, the values are validated
            against the given JSON schema. It is suggested to keep disabled in
            production code as it will slow the command table creation.

            Active validation disabled:

            No validation happens during command table entry modifications, thus
            making the creation of the command table faster.

            Method `is_valid()` can be used for command table validation when active
            validation is disabled. It is recommended to avoid it in production code.

    Example:
    ```
    >>> from zhinst.toolkit import CommandTable
    >>> ct = CommandTable(json_schema)
    ```

    The ``header`` and ``table`` and then be called:

    ```
    >>> ct.header.version
    "1.1"
    >>> ct.header.userString = "My table"
    >>> ct.table[0].amplitude.value = 1
    >>> ct.table[0].amplitude
    1
    >>> ct.as_dict()
    ```

    Example: Active validation

    Using active validation, error raised instantly on incorrect value:

    ```
    >>> ct = CommandTable(json_schema, active_validation=True)
    >>> ct.table[0].amplitude0.value = 999e9
    ValidationError
    ```

    Disabling active validation:

    No ``ValidationError`` is raised during the creation of the command table,
    but once it is uploaded or called ``as_dict()``, the validation happens.

    ```
    >>> ct = CommandTable(json_schema, active_validation=False)
    >>> ct.table[0].amplitude0.value = 999e9  # No errors raised
    >>> ct.as_dict()
    ValidationError
    ```

    Disabling active validation improves the speed of large command tables:

    ```
    >>> for i in range(1024):
    >>>    ct.table[i].waveform.index = 1
    >>>    ct.table[i].amplitude0.value = 1
    >>>    ct.table[i].amplitude1.value = -0.0
    >>>    ct.table[i].amplitude0.increment = False
    >>>    ct.table[i].amplitude0.increment = True
    ```
    """

    def __init__(self, json_schema: t.Union[str, dict], active_validation: bool = True):
        self._ct_schema: dict = _derefence_json(json_schema)
        self._active_validation = active_validation
        self._header: HeaderEntry = self._header_entry()
        self._table: ListEntry = self._table_entry()

    @property
    def active_validation(self) -> bool:
        """State of active validation.

        Returns:
            True if active validation is enabled.
        """
        return self._active_validation

    @active_validation.setter
    def active_validation(self, value: bool) -> None:
        """Active validation.

        Args:
            value: The state of active validation.
        """
        self._active_validation = value
        _change_active_validation(self._table, value)
        _change_active_validation(self._header, value)

    @property
    def header(self) -> HeaderEntry:
        """Header of the built command table.

        Returns:
            Header Entry
        """
        return self._header

    @property
    def table(self) -> ListEntry:
        """Table entry of the built command table.

        Returns:
            List Entry
        """
        return self._table

    def _header_entry(self) -> HeaderEntry:
        return HeaderEntry(
            self._ct_schema["definitions"]["header"],
            ("header",),
            self._ct_schema.get("version", ""),
            self._active_validation,
        )

    def _table_entry(self) -> ListEntry:
        schema = copy.deepcopy(self._ct_schema["definitions"]["table"])
        index_schema = schema["items"]["properties"].pop("index")
        return ListEntry(
            schema=schema,
            index_schema=index_schema,
            path=("table",),
            active_validation=self._active_validation,
        )

    def clear(self) -> None:
        """Clear CommandTable back to its initial state."""
        self._header = self._header_entry()
        self._table = self._table_entry()

    def as_dict(self) -> dict:
        """Return a dictionary representation of the `CommandTable`.

        The function formats the returner value into a schema which is
        accepted by the ZI devices which support command tables.

        The table is validated against the given schema.

        Returns:
            CommandTable as a Python dictionary.

        Raises:
            `~zhinst.toolkit.exceptions.ValidateError`: The command table
                does not correspond to the given JSON schema.
        """
        result = {
            "header": self._header.as_dict(),
            "table": self._table.as_list(),
        }
        if self.active_validation:
            _validate_instance(result, self._ct_schema)
        return result

    def is_valid(self, raise_for_invalid: bool = False) -> bool:
        """Checks if the command table is valid.

        Args:
            raise_for_invalid: Raises exception if the command table is invalid.
                The flag can be used for getting feedback on what is wrong in
                the command table.

        Returns:
            True if the command table is valid.

        Raises:
            ValidationError: If `raise_for_invalid` was set to `True` and the
                command table is invalid.
        """
        # Due to active_validation having a state, we need to force it to True.
        orig_state = self._active_validation
        try:
            self._active_validation = True if self._active_validation is False else True
            self.as_dict()
            self._active_validation = orig_state
            return True
        except ValidationError:
            self._active_validation = orig_state
            if raise_for_invalid:
                raise
            return False

    def update(self, command_table: t.Union[str, dict]) -> None:
        """Update the existing instance of ``CommandTable`` with command table JSON.

        If both command tables have the same properties, the existing ones
        are overwritten by the new command table.

        Args:
            command_table: Existing command table JSON.
        """

        def json_to_dict(json_: t.Union[str, dict]) -> dict:
            if isinstance(json_, str):
                json_ = json.loads(json_)
            return json_  # type: ignore[return-value]

        command_table = json_to_dict(copy.deepcopy(command_table))
        if self._active_validation:
            _validate_instance(command_table, self._ct_schema)

        def build_nodes(path: t.Optional[ParentEntry], index: int, obj: dict) -> None:
            for k, v in obj.items():
                if isinstance(v, dict):
                    build_nodes(getattr(path, k), index, v)
                else:
                    setattr(path, k, v)

        def build_header_nodes(header: HeaderEntry, obj: dict) -> None:
            for k, v in obj.items():
                setattr(header, k, v)

        build_header_nodes(self._header, command_table["header"])

        for item in command_table["table"]:
            index = item.pop("index", None)
            build_nodes(self._table[index], index, item)



"""Module for toolkit representations of native LabOne modules."""

import typing as t

from zhinst.toolkit.driver.modules.base_module import BaseModule
from zhinst.toolkit.driver.modules.daq_module import DAQModule
from zhinst.toolkit.driver.modules.device_settings_module import DeviceSettingsModule
from zhinst.toolkit.driver.modules.impedance_module import ImpedanceModule
from zhinst.toolkit.driver.modules.pid_advisor_module import PIDAdvisorModule
from zhinst.toolkit.driver.modules.precompensation_advisor_module import (
    PrecompensationAdvisorModule,
)
from zhinst.toolkit.driver.modules.scope_module import ScopeModule
from zhinst.toolkit.driver.modules.shfqa_sweeper import SHFQASweeper
from zhinst.toolkit.driver.modules.sweeper_module import SweeperModule

ModuleType = t.Union[
    BaseModule,
    DAQModule,
    SHFQASweeper,
    SweeperModule,
    ScopeModule,
    ImpedanceModule,
    DeviceSettingsModule,
    PIDAdvisorModule,
    PrecompensationAdvisorModule,
]

__all__ = [
    "BaseModule",
    "DAQModule",
    "DeviceSettingsModule",
    "ImpedanceModule",
    "ModuleType",
    "PIDAdvisorModule",
    "PrecompensationAdvisorModule",
    "SHFQASweeper",
    "ScopeModule",
    "SweeperModule",
]



"""PQSC Instrument Driver."""

from __future__ import annotations

from typing import Optional

from zhinst.toolkit.driver.devices.quantum_system_hub import QuantumSystemHub


class PQSC(QuantumSystemHub):
    """High-level driver for the Zurich Instruments PQSC."""

    def arm(
        self,
        *,
        deep=True,
        repetitions: Optional[int] = None,
        holdoff: Optional[float] = None,
    ) -> None:
        """Prepare PQSC for triggering the instruments.

        This method configures the execution engine of the PQSC and
        clears the register bank. Optionally, the *number of triggers*
        and *hold-off time* can be set when specified as keyword
        arguments. If they are not specified, they are not changed.

        Note that the PQSC is disabled at the end of the hold-off time
        after sending out the last trigger. Therefore, the hold-off time
        should be long enough such that the PQSC is still enabled when
        the feedback arrives. Otherwise, the feedback cannot be processed.

        Args:
            deep: A flag that specifies if a synchronization
                should be performed between the device and the data
                server after stopping the PQSC and clearing the
                register bank (default: True).
            repetitions: If specified, the number of triggers sent
                over ZSync ports will be set (default: None).
            holdoff: If specified, the time between repeated
                triggers sent over ZSync ports will be set. It has a
                minimum value and a granularity of 100 ns
                (default: None).

        """
        # Stop the PQSC if it is already running
        self.stop(deep=deep)
        if repetitions is not None:
            self.execution.repetitions(repetitions)
        if holdoff is not None:
            self.execution.holdoff(holdoff)
        # Clear register bank
        self.feedback.registerbank.reset(1, deep=deep)



"""zhinst-toolkit multistate node adaptions."""

import typing as t
from functools import cached_property

import numpy as np

import zhinst.utils.shfqa.multistate as utils
from zhinst.toolkit.nodetree import Node, NodeTree
from zhinst.toolkit.nodetree.helper import create_or_append_set_transaction
from zhinst.toolkit.nodetree.node import NodeList


class Qudit(Node):
    """Single Qudit node.

    Implements basic functionality of a single qudit node, e.g applying the
    basic configuration.

    Args:
        root: Root of the nodetree.
        tree: Tree (node path as tuple) of the current node.
        serial: Serial of the device.
        readout_channel: Index of the readout channel this qudit belongs to.
    """

    def __init__(
        self,
        root: NodeTree,
        tree: tuple,
        serial: str,
        readout_channel: int,
    ):
        super().__init__(root, tree)
        self._daq_server = root.connection
        self._serial = serial
        self._readout_channel = readout_channel

    def configure(
        self,
        qudit_settings: utils.QuditSettings,
        enable: bool = True,
    ) -> None:
        """Compiles a list of transactions to apply the qudit settings to the device.

        Args:
            qudit_settings: The qudit settings to be configured.
            enable: Whether to enable the qudit. (default: True)

        """
        settings = utils.get_settings_transaction(
            self._serial,
            self._readout_channel,
            int(self._tree[-1]),
            qudit_settings,
            enable=enable,
        )
        with create_or_append_set_transaction(self._root):
            for node, value in settings:
                self._root.transaction.add(node, value)


class MultiState(Node):
    """MultiState node.

    Implements basic functionality of the MultiState node.

    Args:
        root: Root of the nodetree.
        tree: Tree (node path as tuple) of the current node.
        serial: Serial of the device.
        index: Index of the corresponding readout channel.
    """

    def __init__(
        self,
        root: NodeTree,
        tree: tuple,
        serial: str,
        index: int,
    ):
        super().__init__(root, tree)
        self._daq_server = root.connection
        self._serial = serial
        self._index = index

    def get_qudits_results(self) -> dict[int, np.ndarray]:
        """Downloads the qudit results from the device and group them by qudit.

        This function accesses the multistate nodes to determine which
        integrators were used for which qudit to able to group the results by
        qudit.

        Returns:
            A dictionary with the qudit index keys and result vector values.
        """
        return utils.get_qudits_results(
            self._daq_server,
            self._serial,
            self._index,
        )

    @cached_property
    def qudits(self) -> t.Sequence[Qudit]:
        """A Sequence of Qudits."""
        return NodeList(
            [
                Qudit(
                    self._root,
                    (*self._tree, "qudits", str(i)),
                    self._serial,
                    self._index,
                )
                for i in range(len(self["qudits"]))
            ],
            self._root,
            (*self._tree, "qudits"),
        )



from unittest.mock import patch

import numpy as np
import pytest

from zhinst.utils.shfqa.multistate import QuditSettings


@pytest.fixture
def multistate(shfqa):
    return shfqa.qachannels[0].readout.multistate


def test_get_qudits_results(mock_connection, multistate):
    with patch("zhinst.toolkit.driver.nodes.multistate.utils", autospec=True) as utils:
        multistate.get_qudits_results()
        utils.get_qudits_results.assert_called_with(
            mock_connection.return_value,
            "DEV1234",
            0,
        )


def test_configure_qudit(mock_connection, multistate):
    with patch("zhinst.toolkit.driver.nodes.multistate.utils", autospec=True) as utils:
        settings = QuditSettings(
            [
                np.random.rand(400),
                np.random.rand(400),
                np.random.rand(400),
                np.random.rand(400),
            ],
        )
        utils.get_settings_transaction.return_value = [
            ("/dev1234/qachannels/0/centerfreq", 1),
            ("/dev1234/qachannels/1/centerfreq", 1),
        ]
        multistate.qudits[0].configure(settings)
        utils.get_settings_transaction.assert_called_with(
            "DEV1234",
            0,
            0,
            settings,
            enable=True,
        )
        mock_connection.return_value.set.assert_called_with(
            utils.get_settings_transaction.return_value,
        )
        # Enable = False
        multistate.qudits[0].configure(settings, enable=False)
        utils.get_settings_transaction.assert_called_with(
            "DEV1234",
            0,
            0,
            settings,
            enable=False,
        )


def test_configure_qudit_existing_transaction(mock_connection, shfqa, multistate):
    with patch("zhinst.toolkit.driver.nodes.multistate.utils", autospec=True) as utils:
        settings = QuditSettings(
            [
                np.random.rand(400),
                np.random.rand(400),
                np.random.rand(400),
                np.random.rand(400),
            ],
        )
        utils.get_settings_transaction.return_value = [
            ("/dev1234/qachannels/0/centerfreq", 1),
            ("/dev1234/qachannels/1/centerfreq", 1),
        ]
        with shfqa.set_transaction():
            shfqa.qachannels[3].centerfreq(1, parse=False)
            multistate.qudits[0].configure(settings)
        utils.get_settings_transaction.assert_called_with(
            "DEV1234",
            0,
            0,
            settings,
            enable=True,
        )
        mock_connection.return_value.set.assert_called_with(
            [("/dev1234/qachannels/3/centerfreq", 1)]
            + utils.get_settings_transaction.return_value,
        )



"""Impedance Module."""

from __future__ import annotations

import logging
import time
import typing as t
from collections.abc import Sequence

from zhinst.core import ImpedanceModule as ZIImpedanceModule
from zhinst.toolkit.driver.modules.base_module import BaseModule

if t.TYPE_CHECKING:  # pragma: no cover
    from zhinst.toolkit.session import Session

logger = logging.getLogger(__name__)


class CalibrationStatus(int, Sequence):
    """Wrapper around a Impedance module status.

    LabOne reports a status for the impedance module as integers.
    The integer needs to be interpreted in a binary format where each
    bit represents a stage within the compensation. If the bit is set it
    means that the step is completed.

    This class wraps around this by both deriving from an integer and a
    Sequence. Therefore one can use it like a int but also access the
    individual steps through items (e.g. module.step[0]).

    Args:
        value (int): Integer value of the status.
    """

    def __new__(cls, value: int):
        """New method of the CalibrationStatus.

        Args:
            value: Integer value of the status.
        """
        new_object = super().__new__(cls, value)
        new_object._value = value  # type: ignore[attr-defined]
        new_object._binary = new_object._to_binary()  # type: ignore[attr-defined]
        return new_object

    def __repr__(self):
        return ", ".join(
            [f"step {i}: {bool(value)}" for i, value in enumerate(self._binary)],
        )

    def _to_binary(self):
        binary = []
        num = self._value
        i = 0
        while num != 0:
            bit = int(num % 2)
            binary.insert(i, bit)
            i = i + 1
            num = int(num / 2)
        return binary

    def __getitem__(self, item):
        return self._binary[item] if len(self._binary) > item else 0

    def __len__(self):
        return len(self._binary)


class ImpedanceModule(BaseModule):
    """Implements a base Impedance Module for Lock-In instruments.

    The Impedance Module corresponds to the Cal sub-tab in the LabOne User
    Interface Impedance Analyzer tab. It allows the user to perform a
    compensation that will be applied to impedance measurements.

    For a complete documentation see the LabOne user manual

    Args:
        impedance_module: Instance of the core Impedance Module.
        session: Session to the Data Server.
    """

    def __init__(self, impedance_module: ZIImpedanceModule, session: Session):
        super().__init__(impedance_module, session)
        self.root.update_nodes(
            {
                "/expectedstatus": {"GetParser": CalibrationStatus},
                "/status": {"GetParser": CalibrationStatus},
            },
            raise_for_invalid_node=False,
        )

    def wait_done(
        self,
        step: t.Optional[int] = None,
        *,
        timeout: float = 20.0,
        sleep_time: float = 0.5,
    ) -> None:
        """Waits until the specified compensation step is complete.

        Args:
            step: The compensation step to wait for completion.
            timeout: The maximum waiting time in seconds for the compensation
                to complete (default: 20).
            sleep_time: Time in seconds to wait between
                requesting the state. (default: 0.5)

        Raises:
            TimeoutError: The compensation is not completed before timeout.
        """
        start_time = time.time()
        while (
            start_time + timeout >= time.time()
            and self.calibrate()
            and not self.finished(step)
        ):
            logger.info(f"Progress: {(self.progress() * 100):.1f}%")
            time.sleep(sleep_time)
        if self.progress() < 1:
            msg = "Impedance module timed out."
            raise TimeoutError(msg)
        if not self.finished(step):
            if step is None:
                msg = (
                    "Impedance module did not reach the status "
                    f"{CalibrationStatus(self.expectedstatus())} that "
                    "corresponds to a full compensation. "
                    f"(current status: {CalibrationStatus(self.status())})"
                )
                raise RuntimeError(
                    msg,
                )
            msg = (
                f"Impedance module did not finish the requested step {step}. "
                f"(current status: {CalibrationStatus(self.status())})"
            )
            raise RuntimeError(
                msg,
            )

    def finish(self) -> None:
        """Stop the module."""
        self._raw_module.finish()

    def finished(self, step: t.Optional[int] = None) -> bool:
        """Check if the calibration or a step of it is finished.

        Args:
            step: Calibration step. If not None this function checks if the
                specified step is finished. Otherwise it checks if the
                hole calibration is done.

        Returns:
            Flag if the calibration or a step is finished.
        """
        if step is None:
            return self.status() == self.expectedstatus()
        return self.status() & (1 << step)



"""Helper functions used in toolkit."""

import re
import typing as t
from collections.abc import Mapping
from contextlib import contextmanager
from functools import wraps

if t.TYPE_CHECKING:  # pragma: no cover
    from zhinst.toolkit.nodetree.node import Node

T = t.TypeVar("T")


class _NodeInfo(t.TypedDict):
    Node: str
    Description: str
    Properties: str
    Type: str
    Unit: str
    Options: dict[str, str]


NodeDoc = dict[str, _NodeInfo]


@contextmanager
def create_or_append_set_transaction(nodetree) -> t.Generator[None, None, None]:
    """Context manager for a transactional set.

    In contrast to the set_transaction from the nodetree this function only
    creates a new transaction if no other is in progress.
    Should only be called withing the toolkit code.

    Warning:
        This function will silently fail if the existing transaction is exited
        before this function finishes.

    Warning:
        The set is always performed as deep set if called on device nodes.

    Example:
        >>> with nodetree.set_transaction():
                nodetree.test[0].a(1)
                with create_or_append_set_transaction(nodetree):
                    nodetree.test[1].a(2)
                nodetree.test[2].a(2)
    """
    if not nodetree.transaction.in_progress():
        with nodetree.set_transaction():
            yield
    else:
        yield


def resolve_wildcards_labone(path: str, nodes: list[str]) -> list[str]:
    """Resolves potential wildcards.

    Also will resolve partial nodes to its leaf nodes.

    Returns:
        List of matched nodes in the raw path format
    """
    node_raw = re.escape(path)
    node_raw = node_raw.replace("/\\*/", "/[^/]*/").replace("/\\*", "/*") + "(/.*)?$"
    node_raw_regex = re.compile(node_raw)
    return list(filter(node_raw_regex.match, nodes))


class NodeDict(Mapping):
    """Mapping of dictionary structure results.

    The mapping allows to access data with both the string and the toolkit
    node objects.

    Args:
        result: A dictionary of node/value pairs.

    Example:
        >>> result = device.demods["*"].enable()
        >>> print(result)
        {
            '/dev1234/demods/0/enable': 0,
            '/dev1234/demods/1/enable': 1,
        }
        >>> result[device.demods[0].enable]
        0
        >>> result["/dev1234/demods/0/enable"]
        0
    """

    def __init__(self, result: dict[str, t.Any]):
        self._result = result

    def __repr__(self):
        return repr(self._result)

    def __getitem__(self, key: t.Union[str, "Node"]):
        return self._result[str(key)]

    def __iter__(self):
        return iter(self._result)

    def __len__(self):
        return len(self._result)

    def to_dict(self) -> dict[str, t.Any]:
        """Convert the WildcardResult to a dictionary.

        After conversion, `Node` objects cannot be used to get items.
        """
        return self._result


def not_callable_in_transactions(
    func: t.Callable[["Node", t.Any], t.Any],
) -> t.Callable[["Node", t.Any], t.Any]:
    """Wrapper to prevent certain functions from being used within a transaction.

    Certain utils functions which that both get and set values would not work like
    expected in a transaction. This wrapper prevents misuse by throwing an error
    in such cases.

    Args:
        func: function to wrap

    Returns:
        Similar function, but not callable from transactions
    """

    @wraps(func)
    def wrapper(node: "Node", *args, **kwargs):
        if node.root.transaction.in_progress():
            msg = f"'{func.__name__}' cannot be called inside a transaction"
            raise RuntimeError(
                msg,
            )
        return func(node, *args, **kwargs)

    return wrapper



"""PQSC and QHub base Instrument Driver."""

from __future__ import annotations

import logging
import time
from typing import Optional, Union

from zhinst.toolkit.driver.devices.base import BaseInstrument
from zhinst.toolkit.exceptions import ToolkitError

logger = logging.getLogger(__name__)


class QuantumSystemHub(BaseInstrument):
    """Base driver for the Zurich Instruments PQSC and QHub.

    This class should not be instantiated directly, but instead through
    PQSC or QHub
    """

    def run(self, *, deep: bool = True) -> None:
        """Start sending out triggers.

        This method activates the trigger generation to trigger all
        connected instruments over ZSync ports.

        Args:
            deep: A flag that specifies if a synchronization
                should be performed between the device and the data
                server after enabling the PQSC/QHub (default: True).

        """
        self.execution.enable(True, deep=deep)

    def arm_and_run(
        self,
        *,
        repetitions: Optional[int] = None,
        holdoff: Optional[float] = None,
    ) -> None:
        """Arm the PQSC/QHub and start sending out triggers.

        Simply combines the methods arm and run. A synchronization
        is performed between the device and the data server after
        arming and running the PQSC/QHub.

        Args:
            repetitions: If specified, the number of triggers sent
                over ZSync ports will be set (default: None).
            holdoff: If specified, the time between repeated
                triggers sent over ZSync ports will be set. It has a
                minimum value and a granularity of 100 ns
                (default: None).

        """
        self.arm(deep=True, repetitions=repetitions, holdoff=holdoff)
        self.run(deep=True)

    def stop(self, *, deep: bool = True) -> None:
        """Stop the trigger generation.

        Args:
            deep: A flag that specifies if a synchronization
                should be performed between the device and the data
                server after disabling the PQSC/QHub (default: True).

        """
        self.execution.enable(False, deep=deep)

    def wait_done(self, *, timeout: float = 10.0, sleep_time: float = 0.005) -> None:
        """Wait until trigger generation and feedback processing is done.

        Args:
            timeout: The maximum waiting time in seconds for the
                PQSC/QHub (default: 10.0).
            sleep_time: Time in seconds to wait between
                requesting PQSC/QHub state

        Raises:
            TimeoutError: If the PQSC/QHub is not done sending out all
                triggers and processing feedback before the timeout.
        """
        try:
            self.execution.enable.wait_for_state_change(
                0,
                timeout=timeout,
                sleep_time=sleep_time,
            )
        except TimeoutError as error:
            msg = f"{self.__class__.__name__:s} timed out."
            raise TimeoutError(msg) from error

    def check_ref_clock(
        self,
        *,
        timeout: float = 30.0,
        sleep_time: float = 1.0,
    ) -> bool:
        """Check if reference clock is locked successfully.

        Args:
            timeout: Maximum time in seconds the program waits
                (default: 30.0).
            sleep_time: Time in seconds to wait between
                requesting the reference clock status (default: 1)

        Raises:
            TimeoutError: If the process of locking to the reference clock
                exceeds the specified timeout.
        """
        ref_clock_status = self.system.clocks.referenceclock.in_.status
        ref_clock = self.system.clocks.referenceclock.in_.source
        ref_clock_actual = self.system.clocks.referenceclock.in_.sourceactual
        try:
            ref_clock_status.wait_for_state_change(
                2,
                invert=True,
                timeout=timeout,
                sleep_time=sleep_time,
            )
        except TimeoutError as error:
            msg = "Timeout during locking to reference clock signal"
            raise TimeoutError(
                msg,
            ) from error
        if ref_clock_status() == 0:
            return True
        if ref_clock_status() == 1 and ref_clock_actual() != ref_clock():
            ref_clock("internal", deep=True)
            logger.error(
                f"There was an error locking the device({self.serial}) "
                f"onto reference clock signal. Automatically switching to internal "
                f"reference clock. Please try again.",
            )
        return False

    def check_zsync_connection(
        self,
        inputs: Union[list[int], int, list[BaseInstrument], BaseInstrument],
        *,
        timeout: float = 10.0,
        sleep_time: float = 0.1,
    ) -> Union[list[bool], bool]:
        """Check if a ZSync connection is established.

        Checks the current status of the instrument connected to the given ports.
        If a instrument(s) is given instead of a port number, first finds the correct
        port number(s).

        Args:
            inputs: The port numbers to check the ZSync connection for.
                It can either be a single port number given as integer, a list
                of several port numbers an instrument or a list of instruments.
            timeout: Maximum time in seconds the program waits (default: 10.0).
            sleep_time: Time in seconds to wait between requesting the reference
                clock status (default: 0.1)

        Raises:
            TimeoutError: If the process of establishing a ZSync connection on
                one of the specified ports exceeds the specified timeout.
        """
        inputs_list = inputs if isinstance(inputs, list) else [inputs]

        start_time = time.time()

        # Check the status of all ports
        status = []
        for input_value in inputs_list:
            # Convert the instrument into a port, if needed
            if isinstance(input_value, BaseInstrument):
                port = self.find_zsync_worker_port(
                    input_value,
                    timeout=max(0, timeout - (time.time() - start_time)),
                    sleep_time=sleep_time,
                )
            else:
                port = input_value

            # Check or wait until the connection is ready
            status.append(
                self._check_zsync_connection(
                    port,
                    timeout=max(0, timeout - (time.time() - start_time)),
                    sleep_time=sleep_time,
                ),
            )
        return status if isinstance(inputs, list) else status[0]

    def _check_zsync_connection(
        self,
        port: int,
        timeout: float,
        sleep_time: float,
    ) -> bool:
        """Check if the ZSync connection on the given port is successful.

        This function checks the current status of the instrument
        connected to the given port.

        Args:
            ports: Port number to check the ZSync connection for.
            timeout: Maximum time in seconds the program waits.
            sleep_time: Time in seconds to wait between requesting the status

        Raises:
            TimeoutError: If the process of establishing a ZSync connection the
                specified port exceeds the specified timeout.
        """
        status_node = self.zsyncs[port].connection.status
        try:
            # Waits until the status node is "connected" (2)
            status_node.wait_for_state_change(2, timeout=timeout, sleep_time=sleep_time)
        except TimeoutError as error:
            status = status_node()
            err_msg = (
                "Timeout while establishing ZSync connection to the instrument "
                f"on the port {port}."
            )

            if status == 0:
                # No connection
                err_msg += "No instrument detected."
            elif status == 1:
                # In progress
                err_msg += (
                    "Connection still in progress. Consider increasing the timeout."
                )
            elif status == 3:
                # Error
                err_msg += (
                    "Impossible to establish a connect. Check cabling and FW version"
                )

            raise TimeoutError(err_msg) from error
        return True

    def find_zsync_worker_port(
        self,
        device: BaseInstrument,
        timeout: float = 10.0,
        sleep_time: float = 0.1,
    ) -> int:
        """Find the ID of the PQSC/QHub ZSync port connected to a given device.

        The function checks until the given timeout for the specified device to
        show up in the connection list.

        Args:
            device: device for which the connected ZSync port shall be found.
            timeout: Maximum time in seconds the program waits (default: 10.0).
            sleep_time: Time in seconds to wait between requesting the port
                serials list (default: 0.1)

        Returns:
            Index of the searched PQSC/QHub ZSync port.

        Raises:
            ToolkitError: If the given device doesn't appear to be connected
                to the PQSC/QHub via ZSync.
        """
        device_serial = device.serial[3:]

        start = time.time()
        while time.time() - start < timeout:
            node_to_serial_dict = self.zsyncs["*"].connection.serial()

            if device_serial in node_to_serial_dict.values():
                break

            time.sleep(sleep_time)
        else:
            msg = (
                f"No ZSync connection found between {self.__class__.__name__:s}"
                f"{self.serial:s} and the device {device.serial:s}."
            )
            raise ToolkitError(
                msg,
            )

        # Get the node of the ZSync connected to the device
        # (will have the form "/devXXXX/zsyncs/N/connection/serial")
        serial_to_node_dict = {
            serial: node for node, serial in node_to_serial_dict.items()
        }
        device_zsync_node = serial_to_node_dict[device_serial]

        # Just interested in knowing N: split in
        # ['', 'devXXXX', 'zsyncs', 'N', 'connection', 'serial']
        # and take fourth value
        return int(device_zsync_node.split("/")[3])



import pytest

from zhinst.toolkit.driver.modules.precompensation_advisor_module import (
    PrecompensationAdvisorModule,
)


@pytest.fixture
def precomensation_advisor_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_precomensation_advisor_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.precompensationAdvisor.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    return PrecompensationAdvisorModule(
        mock_connection.return_value.precompensationAdvisor(),
        session,
    )


def test_repr(precomensation_advisor_module):
    assert "MagicMock(DataServerSession(localhost:8004))" in repr(
        precomensation_advisor_module,
    )


def test_device(zi_devices_json, precomensation_advisor_module, mock_connection):
    connected_devices = "dev1234"

    def get_string_side_effect(arg):
        if arg == "/zi/devices":
            return zi_devices_json
        if arg == "/zi/devices/connected":
            return connected_devices
        if arg == "/dev1234/features/devtype":
            return "Test"
        raise RuntimeError("ZIAPINotFoundException")

    mock_connection.return_value.getString.side_effect = get_string_side_effect

    # get
    mock_connection.return_value.precompensationAdvisor.return_value.getString.return_value = (
        ""
    )
    result = precomensation_advisor_module.device()
    assert result == ""

    mock_connection.return_value.precompensationAdvisor.return_value.getString.return_value = (
        "dev1234"
    )
    result = precomensation_advisor_module.device()
    assert result.serial == "dev1234"

    # set
    precomensation_advisor_module.device("dev1634")
    mock_connection.return_value.precompensationAdvisor.return_value.set.assert_called_with(
        "/device",
        "dev1634",
    )

    precomensation_advisor_module.device(result)
    mock_connection.return_value.precompensationAdvisor.return_value.set.assert_called_with(
        "/device",
        "dev1234",
    )



"""QHub Instrument Driver."""

from __future__ import annotations

from typing import Optional

from zhinst.toolkit.driver.devices.quantum_system_hub import QuantumSystemHub


class QHub(QuantumSystemHub):
    """High-level driver for the Zurich Instruments QHub."""

    def arm(
        self,
        *,
        deep=True,
        repetitions: Optional[int] = None,
        holdoff: Optional[float] = None,
    ) -> None:
        """Prepare QHub for triggering the instruments.

        This method configures the execution engine of QHub.
        Optionally, the *number of triggers*
        and *hold-off time* can be set when specified as keyword
        arguments. If they are not specified, they are not changed.

        Note that the QHub is disabled at the end of the hold-off time
        after sending out the last trigger.

        Args:
            deep: A flag that specifies if a synchronization
                should be performed between the device and the data
                server after stopping QHub (default: True).
            repetitions: If specified, the number of triggers sent
                over ZSync ports will be set (default: None).
            holdoff: If specified, the time between repeated
                triggers sent over ZSync ports will be set. It has a
                minimum value and a granularity of 100 ns
                (default: None).

        """
        # Stop QHub if it is already running
        self.stop(deep=deep)
        if repetitions is not None:
            self.execution.repetitions(repetitions)
        if holdoff is not None:
            self.execution.holdoff(holdoff)



import pathlib
import re

import pytest
import yaml

EXAMPLES_PATH = pathlib.Path("examples")


def match_regex_in_markdown(markdown_file: pathlib.Path, exp_to_match: str) -> bool:
    """Check if a regex is matched inside a markdown file.

    Args:
        markdown_file: Path to the markdown file where the regex is to be searched.
        exp_to_match: Regular expression to be matched.

    Returns:
        Flag if the regex has been matched.
    """
    with open(markdown_file) as f:
        for line in f:
            match = re.search(exp_to_match, line)
            if match is not None:
                return True
    return False


def get_markdown_files() -> pathlib.Path:
    """Collect all the markdown files contained in the examples path.

    Yields:
        The path to a markdown file.

    """
    with open(EXAMPLES_PATH / "test.spec.yml", "rb") as f:
        test_spec = yaml.load(f, Loader=yaml.Loader)

    md_list = EXAMPLES_PATH.glob("*.md")
    for md_file in md_list:
        # README.md should not be checked
        if md_file.stem == "README" or (
            "skip" in test_spec[md_file.stem] and test_spec[md_file.stem]["skip"]
        ):
            continue
        else:
            yield md_file


@pytest.mark.parametrize(
    "input_file",
    get_markdown_files(),
    ids=lambda input_file: input_file.stem,
)
def test_examples_regex_match(input_file: pathlib.Path):
    assert match_regex_in_markdown(
        input_file,
        '"DEVXXXX"',
    ), f'Expression "DEVXXXX" not found in {input_file.name}'

    assert match_regex_in_markdown(
        input_file,
        '"localhost"',
    ), f'Expression "localhost" not found in {input_file.name}'



from zhinst.toolkit import CommandTable, Sequence


def test_setting_command_table(command_table_schema):
    sequencer = Sequence()
    ct = CommandTable(command_table_schema)
    sequencer.command_table = ct
    assert sequencer.command_table == ct


def test_setting_command_table_constructor(command_table_schema):
    ct = CommandTable(command_table_schema)
    sequencer = Sequence(command_table=ct)
    assert sequencer.command_table == ct



"""UHFLI Instrument Driver."""

from __future__ import annotations

import logging
import typing as t
from functools import cached_property

from zhinst.toolkit.driver.devices.base import BaseInstrument
from zhinst.toolkit.driver.nodes.awg import AWG
from zhinst.toolkit.nodetree.node import Node, NodeList

logger = logging.getLogger(__name__)


class UHFLI(BaseInstrument):
    """High-level driver for the Zurich Instruments UHFLI."""

    @cached_property
    def awgs(self) -> t.Union[t.Sequence[AWG], Node]:
        """A Sequence of AWG Cores.

        Device options requirement(s): AWG

        Returns:
            A list of AWG Cores or a single AWG Core node.
        """
        if "AWG" not in self.features.options():
            logger.error("Missing option: AWG")
            return Node(
                self._root,
                (*self._tree, "awgs"),
            )
        return NodeList(
            [
                AWG(
                    self.root,
                    (*self._tree, "awgs", str(i)),
                    self.serial,
                    i,
                    self.device_type,
                    self.device_options,
                )
                for i in range(len(self["awgs"]))
            ],
            self._root,
            (*self._tree, "awgs"),
        )



"""Interface Enums for the zhinst.toolkit."""

from enum import Enum, IntEnum


class SHFQAChannelMode(Enum):
    """SHFQA channel mode.

    Attributes:
        SPECTROSCOPY: Spectroscopy mode.
        READOUT: Readout mode.
    """

    SPECTROSCOPY: str = "spectroscopy"
    READOUT: str = "readout"


class AveragingMode(IntEnum):
    """Averaging modes.

    Attributes:
        CYCLIC: All frequency points are measured once from start frequency to stop
            frequency. The sweeper then moves back to start frequency and repeats the
            sweep the number of times specified by the number of averages setting.
        SEQUENTIAL: A frequency point is measured the number of times specified by the
            number of averages setting. In other words, the same frequency point is
            measured repeatedly until the number of averages is reached and the
            sweeper then moves to the next frequency point.
    """

    CYCLIC: int = 0
    SEQUENTIAL: int = 1



import numpy as np
import pytest

from zhinst.toolkit.driver.modules.pid_advisor_module import PIDAdvisorModule, PIDMode


@pytest.fixture
def pid_advisor_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_pid_advisor_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.pidAdvisor.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    return PIDAdvisorModule(mock_connection.return_value.pidAdvisor(), session)


def test_repr(pid_advisor_module):
    assert "MagicMock(DataServerSession(localhost:8004))" in repr(pid_advisor_module)


def test_pid_mode_node(pid_advisor_module):
    pid_advisor_module.raw_module.getInt.return_value = 0
    assert pid_advisor_module.pid.mode() == PIDMode.NONE

    pid_advisor_module.raw_module.getInt.return_value = 1
    assert pid_advisor_module.pid.mode() == PIDMode.P_Gain

    pid_advisor_module.raw_module.getInt.return_value = 2
    assert pid_advisor_module.pid.mode() == PIDMode.I_Gain

    pid_advisor_module.raw_module.getInt.return_value = 4
    assert pid_advisor_module.pid.mode() == PIDMode.D_Gain

    pid_advisor_module.raw_module.getInt.return_value = 8
    assert pid_advisor_module.pid.mode() == PIDMode.D_Filter_Limit

    pid_advisor_module.raw_module.getInt.return_value = 15
    assert (
        pid_advisor_module.pid.mode()
        == PIDMode.P_Gain | PIDMode.I_Gain | PIDMode.D_Gain | PIDMode.D_Filter_Limit
    )


def test_tuner_mode_node(pid_advisor_module):
    pid_advisor_module.raw_module.getInt.return_value = 0
    assert pid_advisor_module.tuner.mode() == PIDMode.NONE

    pid_advisor_module.raw_module.getInt.return_value = 1
    assert pid_advisor_module.tuner.mode() == PIDMode.P_Gain

    pid_advisor_module.raw_module.getInt.return_value = 2
    assert pid_advisor_module.tuner.mode() == PIDMode.I_Gain

    pid_advisor_module.raw_module.getInt.return_value = 4
    assert pid_advisor_module.tuner.mode() == PIDMode.D_Gain

    pid_advisor_module.raw_module.getInt.return_value = 8
    assert pid_advisor_module.tuner.mode() == PIDMode.D_Filter_Limit

    pid_advisor_module.raw_module.getInt.return_value = 15
    assert (
        pid_advisor_module.tuner.mode()
        == PIDMode.P_Gain | PIDMode.I_Gain | PIDMode.D_Gain | PIDMode.D_Filter_Limit
    )


def test_wait_done(pid_advisor_module):
    pid_advisor_module.raw_module.getInt.return_value = 0
    pid_advisor_module.raw_module.finished.return_value = 1
    pid_advisor_module.raw_module.progress.return_value = np.array([1])
    pid_advisor_module.wait_done()

    pid_advisor_module.raw_module.getInt.return_value = 1
    with pytest.raises(TimeoutError):
        pid_advisor_module.wait_done(timeout=0.01, sleep_time=0.001)

    pid_advisor_module.raw_module.getInt.return_value = 1
    pid_advisor_module.raw_module.progress.return_value = np.array([1])
    with pytest.raises(TimeoutError):
        pid_advisor_module.wait_done(timeout=0.01, sleep_time=0.001)

    pid_advisor_module.raw_module.finished.return_value = 0
    pid_advisor_module.raw_module.progress.return_value = np.array([0.1])
    with pytest.raises(TimeoutError):
        pid_advisor_module.wait_done(timeout=0.01, sleep_time=0.001)



"""zhinst-toolkit AWG node adaptions."""

from __future__ import annotations

import json
import logging
import typing as t
from functools import cached_property

from zhinst.core import compile_seqc
from zhinst.toolkit.driver.nodes.command_table_node import CommandTableNode
from zhinst.toolkit.nodetree import Node, NodeTree
from zhinst.toolkit.nodetree.helper import (
    create_or_append_set_transaction,
)
from zhinst.toolkit.sequence import Sequence
from zhinst.toolkit.waveform import Waveforms

logger = logging.getLogger(__name__)


class AWG(Node):
    """AWG node.

    This class implements the basic functionality for the device specific
    arbitrary waveform generator.
    Besides the upload/compilation of sequences it offers the upload of
    waveforms and command tables.

    Args:
        root: Root of the nodetree
        tree: Tree (node path as tuple) of the current node
        session: Underlying session.
        serial: Serial of the device.
        index: Index of the corresponding awg channel
        device_type: Device type
    """

    def __init__(
        self,
        root: NodeTree,
        tree: tuple,
        serial: str,
        index: int,
        device_type: str,
        device_options: str,
    ):
        Node.__init__(self, root, tree)
        self._daq_server = root.connection
        self._serial = serial
        self._index = index
        self._device_type = device_type
        self._device_options = device_options

    def enable_sequencer(self, *, single: bool) -> None:
        """Starts the sequencer of a specific channel.

        Warning:
            This function is synchronous and blocks until the sequencer is enabled.
            When working with multiple instruments this function is the wrong
            approach and the sequencer should be enabled asynchronously.
            (For more information please take a look at the awg example in the
            toolkit documentation.)

        Args:
            single: Flag if the sequencer should be disabled after finishing
            execution.

        Raises:
            RuntimeError: If the sequencer could not be enabled.
        """
        self.single(single)
        if not self.enable(1, deep=True):
            msg = (
                "The sequencer could not be enabled. Please ensure that the "
                "sequencer program is loaded and configured correctly."
            )
            raise RuntimeError(
                msg,
            )

    def wait_done(self, *, timeout: float = 10, sleep_time: float = 0.005) -> None:
        """Wait until the AWG is finished.

        Args:
            timeout: The maximum waiting time in seconds for the generator
                (default: 10).
            sleep_time: Time in seconds to wait between requesting generator
                state

        Raises:
            RuntimeError: If continuous mode is enabled
            TimeoutError: If the sequencer program did not finish within
                the specified timeout time
        """
        if not self.single():
            msg = (
                f"{self!r}: The generator is running in continuous mode, "
                "it will never be finished."
            )
            raise RuntimeError(
                msg,
            )
        try:
            self.enable.wait_for_state_change(0, timeout=timeout, sleep_time=sleep_time)
        except TimeoutError as error:
            msg = (
                f"{self!r}: The execution of the sequencer program did not finish "
                f"within the specified timeout ({timeout}s)."
            )
            raise TimeoutError(
                msg,
            ) from error

    def compile_sequencer_program(
        self,
        sequencer_program: t.Union[str, Sequence],
        **kwargs: t.Union[str, int],
    ) -> tuple[bytes, dict[str, t.Any]]:
        """Compiles a sequencer program for the specific device.

        Args:
            sequencer_program: The sequencer program to compile.

        Keyword Args:
            samplerate (int): Target sample rate of the sequencer. Only allowed/
                necessary for HDAWG devices. Must correspond to the samplerate
                used by the device (device.system.clocks.sampleclock.freq()).
                If not specified the function will get the value itself from
                the device. It is recommended passing the samplerate if more
                than one sequencer code is uploaded in a row to speed up the
                execution time.
            wavepath (str): path to directory with waveforms. Defaults to path used
                by LabOne UI or AWG Module.
            waveforms (str): waveform CSV files separated by ';'
            output (str): name of embedded ELF filename.

        Returns:
            elf: Binary ELF data for sequencer.
            extra: Extra dictionary with compiler output.

        Example:
            elf, compile_info = device.awgs[0].compile_sequencer_program(seqc)
            device.awgs[0].elf.data(elf)
            device.awgs[0].ready.wait_for_state_change(1)
            device.awgs[0].enable(True)


        Raises:
            RuntimeError: `sequencer_program` is empty.
            RuntimeError: If the compilation failed.
        """
        if "SHFQC" in self._device_type:
            kwargs["sequencer"] = "sg" if "sgchannels" in self._tree else "qa"
        elif "HDAWG" in self._device_type and "samplerate" not in kwargs:
            kwargs["samplerate"] = self.root.system.clocks.sampleclock.freq()

        return compile_seqc(
            str(sequencer_program),
            self._device_type,
            self._device_options,
            self._index,
            **kwargs,
        )

    def load_sequencer_program(
        self,
        sequencer_program: t.Union[str, Sequence],
        **kwargs: t.Union[str, int],
    ) -> dict[str, t.Any]:
        """Compiles the given sequencer program on the AWG Core.

        Warning:
            After uploading the sequencer program one needs to wait before for
            the awg core to become ready before it can be enabled.
            The awg core indicates the ready state through its `ready` node.
            (device.awgs[0].ready() == True)

        Args:
            sequencer_program: Sequencer program to be uploaded.

        Keyword Args:
            samplerate (int): Target sample rate of the sequencer. Only allowed/
                necessary for HDAWG devices. Must correspond to the samplerate
                used by the device (device.system.clocks.sampleclock.freq()).
                If not specified the function will get the value itself from
                the device. It is recommended passing the samplerate if more
                than one sequencer code is uploaded in a row to speed up the
                execution time.
            wavepath (str): path to directory with waveforms. Defaults to path used
                by LabOne UI or AWG Module.
            waveforms (str): waveform CSV files separated by ';'
            output (str): name of embedded ELF filename.

        Example:
            compile_info = device.awgs[0].load_sequencer_program(seqc)
            device.awgs[0].ready.wait_for_state_change(1)
            device.awgs[0].enable(True)

        Raises:
            RuntimeError: `sequencer_program` is empty.
            RuntimeError: If the upload or compilation failed.
        """
        elf, compiler_info = self.compile_sequencer_program(sequencer_program, **kwargs)
        self.elf.data(elf)
        return compiler_info

    def write_to_waveform_memory(
        self,
        waveforms: Waveforms,
        indexes: t.Optional[list] = None,
    ) -> None:
        """Writes waveforms to the waveform memory.

        The waveforms must already be assigned in the sequencer program.

        Args:
            waveforms: Waveforms that should be uploaded.
            indexes: Specify a list of indexes that should be uploaded. If
                nothing is specified all available indexes in waveforms will
                be uploaded. (default = None)
        """
        with create_or_append_set_transaction(self._root):
            for waveform_index in waveforms:
                if indexes and waveform_index not in indexes:
                    continue
                self.root.transaction.add(
                    self.waveform.waves[waveform_index],
                    waveforms.get_raw_vector(waveform_index),
                )

    def read_from_waveform_memory(
        self,
        indexes: t.Optional[list[int]] = None,
    ) -> Waveforms:
        """Read waveforms from the waveform memory.

        Args:
            indexes: List of waveform indexes to read from the device. If not
                specified all assigned waveforms will be downloaded.

        Returns:
            Waveform object with the downloaded waveforms.
        """
        waveform_info = json.loads(self.waveform.descriptors()).get("waveforms", [])
        nodes = [
            self.waveform.node_info.path + f"/waves/{index}"
            for index in range(len(waveform_info))
            if (indexes is None or index in indexes)
            # Entries that have a play_config equals to zero ar dummies/fillers
            # and can therefore be ignored.
            and int(waveform_info[index]["play_config"])
        ]
        nodes_str = ",".join(nodes)
        waveforms_raw = self._daq_server.get(nodes_str, settingsonly=False, flat=True)
        waveforms = Waveforms()
        for node, waveform in waveforms_raw.items():
            slot = int(node.rsplit("/", 1)[-1])
            waveforms.assign_native_awg_waveform(
                slot,
                waveform[0]["vector"],
                channels=int(waveform_info[slot].get("channels", 1)),
                markers_present=bool(int(waveform_info[slot].get("marker_bits")[0])),
            )
        return waveforms

    @cached_property
    def commandtable(self) -> t.Optional[CommandTableNode]:
        """Command table module.

        Returns:
            Command table module.
        """
        if self["commandtable"].is_valid():
            return CommandTableNode(
                self._root,
                (*self._tree, "commandtable"),
                self._device_type,
            )
        return None
